# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Circos(Component):
    """A Circos component.
Dash Circos is a library used to analyze and interpret
data using a circular layout, based on the popular
'Circos' graph. This Dash Bio component is a useful tool
for showcasing relationships bewtween data/datasets in a
beautiful way. Please checkout the Dash Bio repository
on github to learn more about this API.

Keyword arguments:
- enableDownloadSVG (boolean; optional): Allow for an SVG snapshot of the Circos graph to be downloaded.
- enableZoomPan (boolean; optional): Allow for zooming and panning the Circos graph.
- id (string; optional): The ID of the component to be used in Dash callbacks
- style (dict; optional): The CSS styling of the div wrapping the component
- eventDatum (dict; optional): A Dash prop that returns data on clicking or hovering of the tracks.
Depending on what is specified for prop "selectEvent".
- selectEvent (dict; optional): A dictionary used to choose whether tracks should return
data on click, hover, or both, with the dash prop "eventDatum".
The keys of the dictionary represent the index of the list
specified for "tracks".

Ex:
selectEvent={
    "0": "hover",
    "1": "click",
    "2": "both"
    },
- layout (list; required): The overall layout of the Circos graph, provided
as a list of dictionaries.
- config (dict; optional): Configuration of overall layout of the graph.
- size (number; optional): The overall size of the SVG container holding the
graph. Set on initilization and unchangeable thereafter.
- tracks (list; optional): Tracks that specify specific layouts.
For a complete list of tracks and usage,
please check the docs."""
    @_explicitize_args
    def __init__(self, enableDownloadSVG=Component.UNDEFINED, enableZoomPan=Component.UNDEFINED, id=Component.UNDEFINED, style=Component.UNDEFINED, eventDatum=Component.UNDEFINED, selectEvent=Component.UNDEFINED, layout=Component.REQUIRED, config=Component.UNDEFINED, size=Component.UNDEFINED, tracks=Component.UNDEFINED, **kwargs):
        self._prop_names = ['enableDownloadSVG', 'enableZoomPan', 'id', 'style', 'eventDatum', 'selectEvent', 'layout', 'config', 'size', 'tracks']
        self._type = 'Circos'
        self._namespace = 'dash_bio'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['enableDownloadSVG', 'enableZoomPan', 'id', 'style', 'eventDatum', 'selectEvent', 'layout', 'config', 'size', 'tracks']
        self.available_wildcard_properties =            []

        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        for k in ['layout']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Circos, self).__init__(**args)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class OncoPrint(Component):
    """A OncoPrint component.
The OncoPrint component is used to view multiple genetic alteration events
through an interactive and zoomable heatmap. It is a React/Dash port of the
popular oncoPrint() function from the BioConductor R package.
Under the hood, the rendering is done using Plotly.js built upon D3.
Plotly's interactivity allows the user to bind clicks and hovers to genetic
events, allowing the user to create complex bioinformatic apps or workflows
that rely on crossfiltering.
Read more about the component here:
https://github.com/plotly/react-oncoprint

Keyword arguments:
- id (string; optional): The ID of this component, used to identify dash components
in callbacks. The ID needs to be unique to the component.
- eventDatum (dict; optional): A Dash prop that returns data on clicking, hovering or resizing the viewer.
- data (list; optional): Input data, in CBioPortal format where each list entry is a dict
consisting of 'sample', 'gene', 'alteration', and 'type'
- padding (number; optional): Adjusts the padding (as a proportion of whitespace) between two tracks.
Value is a ratio between 0 and 1.
Defaults to 0.05 (i.e., 5 percent). If set to 0, plot will look like a heatmap.
- colorscale (boolean | dict; optional): If not null, will override the default OncoPrint colorscale.
Default OncoPrint colorscale same as CBioPortal implementation.
Make your own colrscale as a {'mutation': COLOR} dict.
Supported mutation keys are ['MISSENSE, 'INFRAME', 'FUSION',
'AMP', 'GAIN', 'HETLOSS', 'HMODEL', 'UP', 'DOWN']
Note that this is NOT a standard plotly colorscale.
- backgroundcolor (string; optional): Default color for the tracks, in common name, hex, rgb or rgba format.
If left blank, will default to a light grey rgb(190, 190, 190).
- range (list; optional): .Toogles whether or not to show a legend on the right side of the plot,
with mutation information.
- showlegend (boolean; optional): .Toogles whether or not to show a legend on the right side of the plot,
with mutation information.
- showoverview (boolean; optional): .Toogles whether or not to show a heatmap overview of the tracks.
- width (number | string; optional): Width of the OncoPrint.
Will disable auto-resizing of plots if set.
- height (number | string; optional): Height of the OncoPrint.
Will disable auto-resizing of plots if set."""
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, eventDatum=Component.UNDEFINED, data=Component.UNDEFINED, padding=Component.UNDEFINED, colorscale=Component.UNDEFINED, backgroundcolor=Component.UNDEFINED, range=Component.UNDEFINED, showlegend=Component.UNDEFINED, showoverview=Component.UNDEFINED, width=Component.UNDEFINED, height=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'eventDatum', 'data', 'padding', 'colorscale', 'backgroundcolor', 'range', 'showlegend', 'showoverview', 'width', 'height']
        self._type = 'OncoPrint'
        self._namespace = 'dash_bio'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'eventDatum', 'data', 'padding', 'colorscale', 'backgroundcolor', 'range', 'showlegend', 'showoverview', 'width', 'height']
        self.available_wildcard_properties =            []

        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        for k in []:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(OncoPrint, self).__init__(**args)



from __future__ import absolute_import

import pandas as pd
from pandas.api.types import is_numeric_dtype
import numpy as np

import plotly.graph_objs as go

GENOMEWIDE_LINE_LABEL = "genomewide_line"
EFFECT_SIZE_LINE_MIN_LABEL = "effect size min line"
EFFECT_SIZE_LINE_MAX_LABEL = "effect size max line"


def _get_hover_text(df, snpname=None, genename=None, annotationname=None):
    """format the hover text used for Manhattan and Volcano Plots
    :param (dataFrame): A pandas dataframe
    :param (string) snpname: A string denoting the column name for the SNP
    names (e.g. rs number). More generally, this column could be anything
    that identifies each point being plotted. For example,
    in an Epigenomewide association study (EWAS) this could be the probe
    name or cg number. This column should be a character. This argument is
    optional, however it is necessary to specify if you want to
    highlight points on the plot using the highlight argument in the
    figure method.
        Default = None
    :param (string) genename: A string denoting the column name for the
    GENE names.
    More generally this could be any annotation information that you
    want to include in the plot.
        Default = None
    :param (string) annotationname: A string denoting the column name
    for an annotation. This could be any annotation information that you
    want to include in the plot (e.g. zscore, effect size, minor allele
    frequency).
        Default = None
    """
    hover_text = ""
    if snpname is not None and snpname in df.columns:
        hover_text = "SNP: " + df[snpname].astype(str)

    if genename is not None and genename in df.columns:
        hover_text = hover_text \
                     + "<br>GENE: " \
                     + df[genename].astype(str)

    if annotationname is not None and annotationname in df.columns:
        hover_text = hover_text \
                     + "<br>" \
                     + df[annotationname].astype(str)
    return hover_text


def VolcanoPlot(
        dataframe,
        effect_size="EFFECTSIZE",
        p="P",
        snp="SNP",
        gene="GENE",
        annotation=None,
        logp=True,
        title="Volcano Plot",
        xlabel=None,
        ylabel="-log10(p)",
        point_size=5,
        col=None,
        effect_size_line=None,
        effect_size_line_color="grey",
        effect_size_line_width=0.5,
        genomewideline_value=-np.log10(5e-8),
        genomewideline_color='grey',
        genomewideline_width=1,
        highlight=True,
        highlight_color="red"
):
    """Returns a figure for a volcano plot.

Keyword arguments:
- dataFrame (dataframe; required): A pandas dataframe which must contain at
    least the following two columns:
            - a numeric quantity to plot such as a p-value or zscore
            - a numeric quantity measuring of the strength of association,
            typically an odds ratio, regression coefficient or log fold
            change. It is referred here as `effect size`.
- p (string; optional): A string denoting the column name for the
    float quantity to be plotted on the y-axis. This column must be
    numeric.  This does not have to be a p-value. It can be any
    numeric quantity such as peak heights, bayes factors, test
    statistics. If it is not a p-value, make sure to set logp =
    FALSE. (Default: "P")
- effect_size (string; optional): A string denoting the column name
    for the effect size. This column in the dataframe must have
    numeric values, with no values being missing nor NaN. (Default:
    "EFFECTSIZE")
- snp (string; optional): A string denoting the column name for the
    SNP names (e.g. rs number). More generally, this column could be
    anything that identifies each point being plotted. For example, in
    an Epigenomewide association study (EWAS) this could be the probe
    name or cg number. This column should be a character. This
    argument is optional, however it is necessary to specify if you
    want to highlight points on the plot using the highlight argument
    in the figure method.(Default: "SNP")
- gene (string; optional): A string denoting the column name for the
    GENE names. More generally, this could be any annotation
    information that should be included in the plot. (Default: "GENE")
- annotation (string; optional): A string denoting the column name for
    an annotation. This could be any annotation information that you
    want to include in the plot (e.g. zscore, effect size, minor
    allele frequency). (Default: None)
- logp (bool; optional): If True, the -log10 of the p-value is
    plotted.  It isn't very useful to plot raw p-values; however,
    plotting the raw value could be useful for other genome-wide plots
    (e.g., peak heights, bayes factors, test statistics, and other
    "scores"). (Default: "True")
- title (string; optional): Title of the graph. (Default: "Volcano
        Plot")
- xlabel (string; optional): Label of the x axis. (Default: None)
- ylabel (string; optional): Label of the y axis. (Default: "-log10(p)")
- point_size (number; optional): Size of the points of the Scatter
  plot. (Default: 5)
- col (string; optional): Color of the points of the Scatter plot. Can
    be in any color format accepted by plotly_js graph_objs. (Default:
    None)
- effect_size_line (bool/list; optional): A boolean which must be
    False to deactivate the option, or a list/array containing the
    upper and lower bounds of the effect size values. Significant data
    points will have lower values than the lower bound, or higher
    values than the higher bound.  Keeping the default value will
    result in assigning the list [-1, 1] to the argument. (Default:
    [-1, 1])
- effect_size_line_color (string; optional): Color of the effect size
    lines. (Default: "grey")
- effect_size_line_width (number; optional): Width of the effect size
  lines. (Default: 2)
- genomewideline_value (bool/number; optional): A boolean which must
    be False to deactivate the option, or a numerical value
    corresponding to the p-value above which the data points are
    considered significant. (Default: -np.log10(5e-8))
- genomewideline_color (string; optional): Color of the genome wide
    line. Can be in any color format accepted by plotly_js
    graph_objs. (Default: "red")
- genomewideline_width (number; optional): Width of the genome wide
  line. (Default: 1)
- highlight (bool; optional): Whether the data points considered
  significant should be highlighted. (Default: True)
- highlight_color (string; optional): Color of the data points
    highlighted because considered as significant Can be in any color
    format accepted by plotly_js graph_objs. (Default: "red")

    # ...
    Example 1: Random Volcano Plot
    '''
    dataframe = pd.DataFrame(
        np.random.randint(0,100,size=(100, 2)),
        columns=['P', 'EFFECTSIZE'])
    fig = create_volcano(dataframe, title='XYZ Volcano plot')

    plotly.offline.plot(fig, image='png')
    '''

    """

    vp = _VolcanoPlot(
        dataframe,
        effect_size=effect_size,
        p=p,
        snp=snp,
        gene=gene,
        annotation=annotation,
        logp=logp
    )

    return vp.figure(
        title=title,
        xlabel=xlabel,
        ylabel=ylabel,
        point_size=point_size,
        col=col,
        effect_size_line=effect_size_line,
        effect_size_line_color=effect_size_line_color,
        effect_size_line_width=effect_size_line_width,
        genomewideline_value=genomewideline_value,
        genomewideline_color=genomewideline_color,
        genomewideline_width=genomewideline_width,
        highlight=highlight,
        highlight_color=highlight_color
    )


class _VolcanoPlot():

    def __init__(
            self,
            x,
            effect_size="EFFECTSIZE",
            p="P",
            snp="SNP",
            gene="GENE",
            annotation=None,
            logp=True
    ):
        """

    Keyword arguments:
    - x (dataframe; required): A pandas dataframe which must contain at
        least the following two columns:
                - a numeric quantity to plot such as a p-value or zscore
                - a numeric quantity measuring of the strength of association,
                typically an odds ratio, regression coefficient or log fold
                change. It is referred here as `effect size`.
    - p (string; optional): A string denoting the column name for the
        float quantity to be plotted on the y-axis. This column must be
        numeric.  This does not have to be a p-value. It can be any
        numeric quantity such as peak heights, bayes factors, test
        statistics. If it is not a p-value, make sure to set logp =
        FALSE. (Default: "P")
    - effect_size (string; optional): A string denoting the column name
        for the effect size. This column in the dataframe must have
        numeric values, with no values being missing nor NaN. (Default:
        "EFFECTSIZE")
    - snp (string; optional): A string denoting the column name for the
        SNP names (e.g. rs number). More generally, this column could be
        anything that identifies each point being plotted. For example, in
        an Epigenomewide association study (EWAS) this could be the probe
        name or cg number. This column should be a character. This
        argument is optional, however it is necessary to specify if you
        want to highlight points on the plot using the highlight argument
        in the figure method.(Default: "SNP")
    - gene (string; optional): A string denoting the column name for the
        GENE names. More generally, this could be any annotation
        information that should be included in the plot. (Default: "GENE")
    - annotation (string; optional): A string denoting the column name for
        an annotation. This could be any annotation information that you
        want to include in the plot (e.g. zscore, effect size, minor
        allele frequency). (Default: None)
    - logp (bool; optional): If True, the -log10 of the p-value is
        plotted.  It isn't very useful to plot raw p-values; however,
        plotting the raw value could be useful for other genome-wide plots
        (e.g., peak heights, bayes factors, test statistics, and other
        "scores"). (Default: "True")

    Returns:
    - object: A Dash Bio ManhattanPlot object."""

        # checking the validity of the arguments

        # Make sure you have effect_size and p columns and that they are of
        # numeric type
        if effect_size not in x.columns.values:
            raise KeyError("Column %s not found in 'x' data.frame"
                           % effect_size)
        else:
            if not is_numeric_dtype(x[effect_size].dtype):
                raise TypeError("%s column should be numeric. Do you have "
                                "'X', 'Y', 'MT', etc? If so change to "
                                "numbers and try again." % effect_size)

        if p not in x.columns.values:
            raise KeyError("Column %s not found in 'x' data.frame" % p)
        else:
            if not is_numeric_dtype(x[p].dtype):
                raise TypeError("%s column should be numeric type" % p)
            else:
                if (x[p] < 0).any():
                    raise ValueError("Negative p-values found."
                                     " These must be removed.")
                if (x[p] > 1).any():
                    raise ValueError("P-values greater than 1 found. "
                                     "These must be removed.")
                if np.isnan(x[p]).any():
                    raise ValueError("NaN P-values found. These must be "
                                     "removed")

        # Create a new DataFrame with columns named after effect_size and p.
        self.data = pd.DataFrame(data=x[[effect_size, p]])

        if snp is not None:
            if snp not in x.columns.values:
                # Warn if you don't have a snp column
                raise KeyError(
                    "snp argument specified as %s but column not found in "
                    "'x' data.frame" % snp)
            else:
                # If the input DataFrame has a snp column, add it to the new
                # DataFrame
                self.data[snp] = x[snp]

        if gene is not None:
            if gene not in x.columns.values:
                # Warn if you don't have a gene column
                raise KeyError(
                    "gene argument specified as %s but column not found in "
                    "'x' data.frame" % gene)
            else:
                # If the input DataFrame has a gene column, add it to the new
                # DataFrame
                self.data[gene] = x[gene]

        if annotation is not None:
            if annotation not in x.columns.values:
                # Warn if you don't have an annotation column
                raise KeyError(
                    "annotation argument specified as %s but column not "
                    "found in 'x' data.frame" % annotation
                )
            else:
                # If the input DataFrame has a gene column, add it to the new
                # DataFrame
                self.data[annotation] = x[annotation]

        self.xlabel = "Effect Size"
        self.ticks = []
        self.ticksLabels = []
        self.effectSize = effect_size
        self.pName = p
        self.snpName = snp
        self.geneName = gene
        self.annotationName = annotation
        self.logp = logp

    def figure(
            self,
            title="Volcano Plot",
            xlabel=None,
            ylabel='-log10(p)',
            point_size=5,
            col=None,
            effect_size_line=None,
            effect_size_line_color="grey",
            effect_size_line_width=0.5,
            genomewideline_value=-np.log10(5e-8),
            genomewideline_color='grey',
            genomewideline_width=1,
            highlight=True,
            highlight_color="red",
    ):
        """

    Keyword arguments:
    - title (string; optional): Title of the graph. (Default: "Volcano
    Plot")
    - xlabel (string; optional): Label of the x axis. (Default: None)
    - ylabel (string; optional): Label of the y axis. (Default: "-log10(p)")
    - point_size (number; optional): Size of the points of the Scatter
      plot. (Default: 5)
    - col (string; optional): Color of the points of the Scatter plot. Can
        be in any color format accepted by plotly_js graph_objs. (Default:
        None)
    - effect_size_line (bool/list; optional): A boolean which must be
        False to deactivate the option, or a list/array containing the
        upper and lower bounds of the effect size values. Significant data
        points will have lower values than the lower bound, or higher
        values than the higher bound.  Keeping the default value will
        result in assigning the list [-1, 1] to the argument. (Default:
        [-1, 1])
    - effect_size_line_color (string; optional): Color of the effect size
        lines. (Default: "grey")
    - effect_size_line_width (number; optional): Width of the effect size
      lines. (Default: 2)
    - genomewideline_value (bool/number; optional): A boolean which must
        be False to deactivate the option, or a numerical value
        corresponding to the p-value above which the data points are
        considered significant. (Default: -np.log10(5e-8))
    - genomewideline_color (string; optional): Color of the genome wide
        line. Can be in any color format accepted by plotly_js
        graph_objs. (Default: "red")
    - genomewideline_width (number; optional): Width of the genome wide
      line. (Default: 1)
    - highlight (bool; optional): Whether the data points considered
      significant should be highlighted. (Default: True)
    - highlight_color (string; optional): Color of the data points
        highlighted because considered as significant Can be in any color
        format accepted by plotly_js graph_objs. (Default: "red")

    Returns:
    - object: A figure compatible with plotly.graph_objs."""

        if xlabel is None:
            xlabel = self.xlabel

        if effect_size_line is None:
            effect_size_line = [-1, 1]

        if not effect_size_line and not isinstance(effect_size_line, bool):
            raise ValueError("If effect_size_line is a logical, it must be "
                             "set to FALSE")

        if np.size(effect_size_line) > 2:
            raise ValueError("The argument effect_size_line should be a "
                             "vector or a list of maximum two components")

        # Initialize plot
        xmin = min(self.data[self.effectSize].values)
        xmax = max(self.data[self.effectSize].values)
        # Taking 105% of the max value of data for x axis range
        xlim = 1.05 * np.max(np.abs([xmin, xmax]))

        if self.logp:
            ymin = min(-np.log10(self.data[self.pName].values))
            ymax = max(-np.log10(self.data[self.pName].values))
        else:
            ymin = min(self.data[self.pName].values)
            ymax = max(self.data[self.pName].values)

        if col is None:
            col = 'black'

        layout = go.Layout(
            title=title,
            hovermode='closest',
            legend={
                'x': 0.85,
                'y': 0.1,
                'bgcolor': '#f2f5fa'
            },
            xaxis={
                'title': xlabel,
                'zeroline': False,
                'range': [-xlim, xlim]
            },
            yaxis={
                'title': ylabel,
                'zeroline': False
            }
        )

        data_to_plot = []  # To contain the data traces
        tmp = pd.DataFrame()  # Empty DataFrame to contain the highlighted data

        if highlight:
            if not isinstance(highlight, bool):
                if self.snpName not in self.data.columns.values:
                    raise KeyError(
                        "snp argument specified for highlight as %s but "
                        "column not found in the data.frame" % self.snpName
                    )
            else:
                if not genomewideline_value:
                    raise Warning(
                        "The genomewideline_value you entered is not a "
                        "positive value, or False, you cannot set highlight "
                        "to True in that case.")
                tmp = self.data

                # Sort the p-values (or -log10(p-values) above the line
                if genomewideline_value:
                    if self.logp:
                        tmp = tmp.loc[-np.log10(tmp[self.pName])
                                      > genomewideline_value]
                    else:
                        tmp = tmp.loc[tmp[self.pName] > genomewideline_value]

                # Sort the effect size in large positive and large negative
                if effect_size_line:
                    lp = tmp.loc[tmp[self.effectSize] > max(effect_size_line)]
                    ln = tmp.loc[tmp[self.effectSize] < min(effect_size_line)]
                    tmp = pd.concat([lp, ln])

                highlight_hover_text = _get_hover_text(
                    tmp,
                    snpname=self.snpName,
                    genename=self.geneName,
                    annotationname=self.annotationName
                )

                if not tmp.empty:
                    data_to_plot.append(
                        go.Scattergl(
                            x=tmp[self.effectSize],
                            y=-np.log10(tmp[self.pName].values) if self.logp
                            else tmp[self.pName].values,
                            mode="markers",
                            text=highlight_hover_text,
                            marker=dict(
                                color=highlight_color,
                                size=point_size),
                            name="Point(s) of interest"
                        )
                    )

        # Remove the highlighted data from the DataFrame if not empty
        if tmp.empty:
            data = self.data
        else:
            data = self.data.drop(self.data.index[tmp.index])

        hover_text = _get_hover_text(
            data,
            snpname=self.snpName,
            genename=self.geneName,
            annotationname=self.annotationName
        )

        data_to_plot.append(
            go.Scattergl(
                x=data[self.effectSize].values,
                y=-np.log10(data[self.pName].values) if self.logp
                else data[self.pName].values,
                mode="markers",
                marker={
                    'color': col,
                    'size': point_size,
                    # 'name': "chr%i" % self.data[self.chrName].unique()
                },
                text=hover_text,
                name="Dataset"
            )
        )

        # Draw the effect size lines
        if effect_size_line:
            lines = [
                go.layout.Shape(
                    name=EFFECT_SIZE_LINE_MIN_LABEL,
                    type="line",
                    line=dict(
                        color=effect_size_line_color,
                        width=effect_size_line_width,
                        dash='dash'
                    ),
                    x0=effect_size_line[0], x1=effect_size_line[0], xref="x",
                    y0=ymin, y1=ymax, yref="y"
                ),
                go.layout.Shape(
                    name=EFFECT_SIZE_LINE_MAX_LABEL,
                    type="line",
                    line=dict(
                        color=effect_size_line_color,
                        width=effect_size_line_width,
                        dash='dash'
                    ),
                    x0=effect_size_line[1], x1=effect_size_line[1], xref="x",
                    y0=ymin, y1=ymax, yref="y"
                ),
            ]
        else:
            lines = []

        if genomewideline_value:
            genomewideline = go.layout.Shape(
                name=GENOMEWIDE_LINE_LABEL,
                type="line",
                line=dict(
                    color=genomewideline_color,
                    width=genomewideline_width,
                    dash='dash'
                ),
                x0=-xlim, x1=xlim, xref="x",
                y0=genomewideline_value, y1=genomewideline_value, yref="y"
            )
            lines.append(genomewideline)

        layout.shapes = lines

        return go.Figure(data=data_to_plot, layout=layout)



# -*- coding: utf-8 -*-
from random import shuffle

import numpy as np
import scipy as scp
import scipy.cluster.hierarchy as sch
import scipy.spatial as scs
from sklearn.impute import SimpleImputer

import plotly.graph_objs as go
from plotly import tools


# pylint: disable=assignment-from-no-return, no-self-use
def Clustergram(
        data=None,
        generate_curves_dict=False,
        return_computed_traces=False,
        computed_traces=None,
        row_labels=None,
        column_labels=None,
        hide_labels=None,
        standardize='none',
        cluster='all',
        row_dist='euclidean',
        col_dist='euclidean',
        dist_fun=scs.distance.pdist,
        link_fun=lambda x, **kwargs: sch.linkage(x, 'complete', **kwargs),
        color_threshold=None,
        optimal_leaf_order=False,
        color_map=None,
        color_list=None,
        display_range=3,
        symmetric_value=True,
        log_transform=False,
        display_ratio=0.2,
        imputer_parameters=None,
        row_group_marker=None,    # group number, annotation, color
        col_group_marker=None,    # same as above
        tick_font=None,
        annotation_font=None,
        line_width=0.5,
        paper_bg_color='rgba(0,0,0,0)',
        plot_bg_color='rgba(0,0,0,0)',
        height=500,
        width=500
):
    """Function that returns a Dash Bio Clustergram object.

Keyword arguments:

- data (ndarray; required): Matrix of observations as array of arrays
- generate_curves_dict (bool; optional): Whether or not to return a
    dictionary containing information about the cluster number
    associated with each curve number in the graph. (May be useful
    if one wishes to capture the cluster number that is clicked.)
    (Default: False)
- return_computed_traces (bool; optional): Whether or not to return
    the precomputed dendrogram traces. (May be useful if one wishes
    to add, e.g., group markers to the figure without recalculating
    the clustering in the entire figure.) (Default: False)
- computed_traces (dict; optional): The dendrogram traces from another
   Clustergram component.
- row_labels (list; optional): List of row category labels
   (observation labels)
- column_labels (list; optional): List of column category labels
   (observation labels)
- hide_labels (list; optional): List of labels not to display on the
    final plot.
- standardize (string; optional): The dimension for standardizing
    values, so that the mean is 0 and the standard deviation is 1
    along the specified dimension: 'row', 'column', or 'none'.
    (Default: 'none')
- cluster (string; optional): The dimension along which the data will
    be clustered: 'row', 'column', or 'all'; 'all' means data to be
    clustered along columns, then clustered along rows of
    row-clustered data. (Default: 'all')
- row_dist (function; optional): Function specifying the distance
    metric for rows that will be passed to the function specified in
    dist_fun (see scipy.spatial.distance.pdist). (Default:
    'euclidean')
- col_dist (function; optional): Function specifying the distance
    metric for columns that will be passed to the function specified
    in dist_fun (see scipy.spatial.distance.pdist). (Default:
    'euclidean')
- dist_fun (function; optional): Function to compute the pairwise
    distance from the observations (see
    scipy.spatial.distance.pdist). (Default: scs.distance.pdist)
- link_fun (function; optional): Function to compute the linkage
    matrix from the pairwise distances (see
    scipy.cluster.hierarchy.linkage). (Default: sch.linkage)
- color_threshold (dict; optional): Maximum linkage value for which
    unique colors are assigned to clusters; 'row' for rows, and 'col'
    for columns. (Default: {'row': 0, 'col': 0})
- optimal_leaf_order (bool; optional): Enabling/disabling of the
    option to determine leaf order that maximizes similarity between
    neighboring leaves. (Default: False)
- color_map (list; optional): The colorscale for the heatmap. Each
    list element contains two elements; the first element refers to
    the portion of the maximum data point under which a cell will be
    colored, and the second element refers to the color. e.g., a
    colorscale [[0.0, 'white'], [0.5, 'gray'], [1.0, 'black']] mean
    that for all cells with a value in the 50th or lower percentile of
    the dataset, the color on the heatmap would be white; all cells
    with a value in the 50th or higher percentile, excluding the 100th
    percentile, would be gray; and the cell(s) in the 100th percentile
    would be colored black. (Default: [[0.0, 'rgb(255,0,0)'], [0.5,
    'rgb(0,0,0)'], [1.0, 'rgb(0,255,0)']])
- color_list (dict; optional): The list of colors to use for different
   clusters in the dendrogram that have a root under the threshold for
   each dimension. If there are fewer colors than there are clusters
   along a specific dimension. The keys are: 'row' (for row clusters),
   'col' (for column clusters), and 'bg' (for all traces above the
   clustering threshold for both row and column. (Default: None)
- display_range (double; optional): In the heatmap, standardized
    values from the dataset that are below the negative of this value
    will be colored with one shade, and the values that are above this
    value will be colored with another. (Default: 3)
- symmetric_value (bool; optional): Whether or not to center the
    values of the heatmap about zero. (Default: True)
- log_transform (bool; optional): Whether or not to transforms the
    data by taking the base-two logarithm of all values in the
    dataset. (Default: False)
- display_ratio (list/float; optional): The dendrograms' heights with
    respect to the size of the heatmap; with one element, both the row
    and column dendrograms have the same ratio; with two, the row
    dendrogram ratio corresponds to the first element of the list and
    the column dendrogram ratio corresponds to the second element of
    the list. (Default: 0.2)
- imputer_parameters (dict; optional): Specifies the parameters
    'missing_values' and 'strategy' of the SimpleImputer class from
    scikit-learn 0.20.1 (both of these parameters must be keys in the
    dictionary).  An additional parameter, 'axis', is used to specify
    the direction along which to impute (a parameter of Imputer, which
    was deprecated in scikit-learn 0.20.0).  'axis=0' indicates that
    imputing should happen along columns, while 'axis=1' indicates
    that it should happen along rows (see: https://scikit
    -learn.org/stable/modules/generated/sklearn.preprocessing.Imputer.html).
    (Default: None)
- row_group_marker (list; optional) A list containing the annotations
    for row clusters in the dendrogram. Each annotation is a
    dictionary with the keys 'group_number' (the cluster number to
    highlight), 'annotation' (a string containing the text of the
    annotation), and 'color' (a string representation of the color of
    the annotation). (Default: None)
- col_group_marker (list; optional): A list containing the annotations for
    column clusters in the dendrogram. Each annotation is a dictionary
    with the keys 'group_number' (the cluster number to highlight),
    'annotation' (a string containing the text of the annotation), and
    'color' (a string representation of the color of the
    annotation). (Default: None)
- tick_font (dict; optional): The font options for ticks, as specified
   in the Plotly graph_objs documentation (see:
   https://plot.ly/python/reference/#bar-marker-colorbar-tickfont). (Default:
   None)
- annotation_font (dict; optional): The font options for annotations,
   as specified in the Plotly graph_objs documentation (see:
   https://plot.ly/python/reference/#layout-scene-annotations-items-annotation-font). (Default:
   None)
- line_width (list/float; optional): The line width for the
    dendrograms. If in list format, the first element corresponds to
    the width of the row dendrogram traces, and the second corresponds
    to the width of the column dendrogram traces. (Default: 0.5)
- paper_bg_color (string; optional): The background color of the paper on the
    graph. (Default: 'rgba(0,0,0,0)')
- plot_bg_color (string; optional): The background color of the subplots on the
    graph. (Default: 'rgba(0,0,0,0)')
- height (int; optional): The height of the graph, in px. (Default:
  500)
- width (int; optional): The width of the graph, in px. (Default: 500)"""
    if hide_labels is None:
        hide_labels = []
    if color_threshold is None:
        color_threshold = dict(row=0, col=0)

    # get rid of arguments that are not used by _Clustergram
    kwargs = locals()
    kwargs.pop('return_computed_traces')
    kwargs.pop('computed_traces')
    kwargs.pop('generate_curves_dict')

    (fig, ct, curves_dict) = _Clustergram(
        **kwargs
    ).figure(
        computed_traces=computed_traces
    )

    return_values = [go.Figure(fig)]

    if generate_curves_dict:
        return_values.append(curves_dict)
    if return_computed_traces:
        return_values.append(ct)

    # return only the figure by default
    if len(return_values) == 1:
        return return_values[0]

    # otherwise, return all requested values
    return tuple(return_values)


class _Clustergram():
    """A Dash Bio Clustergram class.

Keyword arguments:

- data (ndarray; required): Matrix of observations as array of arrays
- row_labels (list; optional): List of row category labels
   (observation labels)
- column_labels (list; optional): List of column category labels
   (observation labels)
- hide_labels (list; optional): List of labels not to display on the
    final plot.
- standardize (string; optional): The dimension for standardizing
    values, so that the mean is 0 and the standard deviation is 1
    along the specified dimension: 'row', 'column', or 'none'.
    (Default: 'none')
- cluster (string; optional): The dimension along which the data will
    be clustered: 'row', 'column', or 'all'; 'all' means data to be
    clustered along columns, then clustered along rows of
    row-clustered data. (Default: 'all')
- row_dist (function; optional): Function specifying the distance
    metric for rows that will be passed to the function specified in
    dist_fun (see scipy.spatial.distance.pdist). (Default:
    'euclidean')
- col_dist (function; optional): Function specifying the distance
    metric for columns that will be passed to the function specified
    in dist_fun (see scipy.spatial.distance.pdist). (Default:
    'euclidean')
- dist_fun (function; optional): Function to compute the pairwise
    distance from the observations (see
    scipy.spatial.distance.pdist). (Default: scs.distance.pdist)
- link_fun (function; optional): Function to compute the linkage
    matrix from the pairwise distances (see
    scipy.cluster.hierarchy.linkage). (Default: sch.linkage)
- color_threshold (dict; optional): Maximum linkage value for which
    unique colors are assigned to clusters; 'row' for rows, and 'col'
    for columns. (Default: {'row': 0, 'col': 0})
- optimal_leaf_order (bool; optional): Enabling/disabling of the
    option to determine leaf order that maximizes similarity between
    neighboring leaves. (Default: False)
- color_map (list; optional): The colorscale for the heatmap. Each
    list element contains two elements; the first element refers to
    the portion of the maximum data point under which a cell will be
    colored, and the second element refers to the color. e.g., a
    colorscale [[0.0, 'white'], [0.5, 'gray'], [1.0, 'black']] mean
    that for all cells with a value in the 50th or lower percentile of
    the dataset, the color on the heatmap would be white; all cells
    with a value in the 50th or higher percentile, excluding the 100th
    percentile, would be gray; and the cell(s) in the 100th percentile
    would be colored black. (Default: [[0.0, 'rgb(255,0,0)'], [0.5,
    'rgb(0,0,0)'], [1.0, 'rgb(0,255,0)']])
- color_list (dict; optional): The list of colors to use for different
   clusters in the dendrogram that have a root under the threshold for
   each dimension. If there are fewer colors than there are clusters
   along a specific dimension. The keys are: 'row' (for row clusters),
   'col' (for column clusters), and 'bg' (for all traces above the
   clustering threshold for both row and column. (Default: None)
- display_range (double; optional): In the heatmap, standardized
    values from the dataset that are below the negative of this value
    will be colored with one shade, and the values that are above this
    value will be colored with another. (Default: 3)
- symmetric_value (bool; optional): Whether or not to center the
    values of the heatmap about zero. (Default: True)
- log_transform (bool; optional): Whether or not to transforms the
    data by taking the base-two logarithm of all values in the
    dataset. (Default: False)
- display_ratio (list/float; optional): The dendrograms' heights with
    respect to the size of the heatmap; with one element, both the row
    and column dendrograms have the same ratio; with two, the row
    dendrogram ratio corresponds to the first element of the list and
    the column dendrogram ratio corresponds to the second element of
    the list. (Default: 0.2)
- imputer_parameters (dict; optional): Specifies the parameters
    'missing_values' and 'strategy' of the SimpleImputer class from
    scikit-learn 0.20.1 (both of these parameters must be keys in the
    dictionary).  An additional parameter, 'axis', is used to specify
    the direction along which to impute (a parameter of Imputer, which
    was deprecated in scikit-learn 0.20.0).  'axis=0' indicates that
    imputing should happen along columns, while 'axis=1' indicates
    that it should happen along rows (see: https://scikit
    -learn.org/stable/modules/generated/sklearn.preprocessing.Imputer.html).
    (Default: None)
- row_group_marker (list; optional) A list containing the annotations
    for row clusters in the dendrogram. Each annotation is a
    dictionary with the keys 'group_number' (the cluster number to
    highlight), 'annotation' (a string containing the text of the
    annotation), and 'color' (a string representation of the color of
    the annotation). (Default: None)
- col_group_marker (list; optional): A list containing the annotations for
    column clusters in the dendrogram. Each annotation is a dictionary
    with the keys 'group_number' (the cluster number to highlight),
    'annotation' (a string containing the text of the annotation), and
    'color' (a string representation of the color of the
    annotation). (Default: None)
- tick_font (dict; optional): The font options for ticks, as specified
   in the Plotly graph_objs documentation (see:
   https://plot.ly/python/reference/#bar-marker-colorbar-tickfont). (Default:
   None)
- annotation_font (dict; optional): The font options for annotations,
   as specified in the Plotly graph_objs documentation (see:
   https://plot.ly/python/reference/#layout-scene-annotations-items-annotation-font). (Default:
   None)
- line_width (list/float; optional): The line width for the
    dendrograms. If in list format, the first element corresponds to
    the width of the row dendrogram traces, and the second corresponds
    to the width of the column dendrogram traces. (Default: 0.5)
- paper_bg_color (string; optional): The background color of the paper on the
    graph. (Default: 'rgba(0,0,0,0)')
- plot_bg_color (string; optional): The background color of the subplots on the
    graph. (Default: 'rgba(0,0,0,0)')
- height (int; optional): The height of the graph, in px. (Default:
  500)
- width (int; optional): The width of the graph, in px. (Default: 500)"""

    def __init__(
            self,
            data=None,
            row_labels=None,
            column_labels=None,
            hide_labels=None,
            standardize='none',
            cluster='all',
            row_dist='euclidean',
            col_dist='euclidean',
            dist_fun=scs.distance.pdist,
            link_fun=lambda x, **kwargs: sch.linkage(x, 'complete', **kwargs),
            color_threshold=None,
            optimal_leaf_order=False,
            color_map=None,
            color_list=None,
            display_range=3,
            symmetric_value=True,
            log_transform=False,
            display_ratio=0.2,
            imputer_parameters=None,
            row_group_marker=None,    # group number, annotation, color
            col_group_marker=None,    # same as above
            tick_font=None,
            annotation_font=None,
            line_width=0.5,
            paper_bg_color='rgba(0,0,0,0)',
            plot_bg_color='rgba(0,0,0,0)',
            height=500,
            width=500
    ):
        if hide_labels is None:
            hide_labels = []
        if color_threshold is None:
            color_threshold = dict(row=0, col=0)
        if row_labels is None:
            row_labels = [str(i) for i in range(data.shape[0])]
            hide_labels.append('row')
        if column_labels is None:
            column_labels = [str(i) for i in range(data.shape[1])]
            hide_labels.append('col')

        self._data = data
        self._row_labels = row_labels
        self._column_labels = column_labels
        self._cluster = cluster
        self._row_dist = row_dist
        self._col_dist = col_dist
        self._dist_fun = dist_fun
        self._link_fun = link_fun
        self._color_threshold = color_threshold
        self._optimal_leaf_order = optimal_leaf_order
        if color_map is None:
            self._color_map = [[0.0, 'rgb(255,0,0)'],
                               [0.5, 'rgb(0,0,0)'],
                               [1.0, 'rgb(0,255,0)']]
        else:
            self._color_map = color_map
        self._color_list = color_list
        self._display_range = display_range
        self._symmetric_value = symmetric_value
        self._display_ratio = display_ratio
        self._imputer_parameters = imputer_parameters
        if row_group_marker is None:
            self._row_group_marker = []
        else:
            self._row_group_marker = row_group_marker
        if col_group_marker is None:
            self._col_group_marker = []
        else:
            self._col_group_marker = col_group_marker
        if tick_font is None:
            self._tick_font = dict()
        else:
            self._tick_font = tick_font
        if annotation_font is None:
            self._annotation_font = dict()
        else:
            self._annotation_font = annotation_font
        self._paper_bg_color = paper_bg_color
        self._plot_bg_color = plot_bg_color
        self._height = height
        self._width = width

        # convert line width to list if necessary
        self._line_width = [0, 0]
        if isinstance(line_width, list):
            self._line_width = line_width
        else:
            self._line_width = [line_width, line_width]

        # convert display ratio to list if necessary
        if not isinstance(display_ratio, list):
            self._display_ratio = [display_ratio, display_ratio]
        if self._cluster == 'row':
            self._display_ratio = [display_ratio[0], 0]
        elif self._cluster == 'col':
            self._display_ratio = [0, display_ratio[1]]

        self._hide_labels = []

        if 'row' in hide_labels:
            self._hide_labels.append('yaxis5')
        if 'col' in hide_labels:
            self._hide_labels.append('xaxis5')

        # preprocessing data
        if self._imputer_parameters is not None:

            # numpy NaN values are not serializable and turn into
            # 'None' by the time they get here; passing a string
            # means that it can be converted in the clustergram
            # component itself
            if self._imputer_parameters['missing_values'].lower() == 'nan':
                self._imputer_parameters.update(
                    missing_values=np.nan
                )

            imp = SimpleImputer(
                missing_values=self._imputer_parameters['missing_values'],
                strategy=self._imputer_parameters['strategy']
            )

            if self._imputer_parameters['axis'] == 0:
                self._data = imp.fit_transform(self._data.T).T
            else:
                self._data = imp.fit_transform(self._data)

        if log_transform:
            self._data = np.log2(self._data)
        if standardize in ['row', 'column']:
            self._data = self._scale(self._data, standardize)

    def figure(
            self,
            computed_traces=None
    ):
        # use, if available, the precomputed dendrogram and heatmap
        # traces (as well as the row and column labels)
        dt, heatmap = None, None

        if computed_traces is None:
            dt, self._data, \
                self._row_labels, self._column_labels = self._compute_clustered_data()
        else:
            dt = computed_traces['dendro_traces']
            heatmap = computed_traces['heatmap']
            self._row_labels = computed_traces['row_labels']
            self._column_labels = computed_traces['column_labels']

        # this dictionary relates curve numbers (accessible from the
        # hoverData/clickData props) to cluster numbers
        cluster_curve_numbers = {}

        # initialize plot; GM is for group markers
        # [empty]      [col. dendro] [col. dendro] [empty]
        # [row dendro] [heatmap]     [heatmap]     [row GM]
        # [row dendro] [heatmap]     [heatmap]     [row GM]
        # [empty]      [col. GM]     [col. GM]     [empty]
        fig = tools.make_subplots(
            rows=4, cols=4,
            specs=[
                [{}, {'colspan': 2}, None, {}],
                [{'rowspan': 2}, {'colspan': 2, 'rowspan': 2},
                 None, {'rowspan': 2}],
                [None, None, None, None],
                [{}, {'colspan': 2}, None, {}]
            ],
            vertical_spacing=0,
            horizontal_spacing=0,
            print_grid=False
        )

        fig['layout'].update(
            hovermode='closest'
        )

        # get the tick values; these will be at the leaves of the
        # dendrogram

        tickvals_col = []
        tickvals_row = []

        # for column dendrogram, leaves are at bottom (y=0)
        for i in range(len(dt['col'])):
            xs = dt['col'][i]['x']
            ys = dt['col'][i]['y']

            # during serialization (e.g., in a dcc.Store, the NaN
            # values become None and the arrays get turned into lists;
            # they must be converted back
            if isinstance(xs, list):
                xs = np.array(xs, dtype=np.float)
                dt['col'][i].update(
                    x=xs
                )
            if isinstance(ys, list):
                ys = np.array(ys, dtype=np.float)
                dt['col'][i].update(
                    y=ys
                )
            tickvals_col += [
                xs.flatten()[j]
                for j in range(len(xs.flatten()))
                if ys.flatten()[j] == 0.0 and
                xs.flatten()[j] % 10 == 5
            ]
        tickvals_col = list(set(tickvals_col))

        # for row dendrogram, leaves are at right(x=0, since we
        # horizontally flipped it)
        for i in range(len(dt['row'])):
            xs = dt['row'][i]['x']
            ys = dt['row'][i]['y']

            if isinstance(xs, list):
                xs = np.array(xs, dtype=np.float)
                dt['row'][i].update(
                    x=xs
                )
            if isinstance(ys, list):
                ys = np.array(ys, dtype=np.float)
                dt['row'][i].update(
                    y=ys
                )

            tickvals_row += [
                ys.flatten()[j]
                for j in range(len(ys.flatten()))
                if xs.flatten()[j] == 0.0 and
                ys.flatten()[j] % 10 == 5
            ]

        tickvals_row = list(set(tickvals_row))

        # sort so they are in the right order (lowest to highest)
        tickvals_col.sort()
        tickvals_row.sort()

        # update axis settings for dendrograms and heatmap
        axes = ['xaxis1', 'xaxis2', 'xaxis4', 'xaxis5',
                'yaxis1', 'yaxis2', 'yaxis4', 'yaxis5']

        for a in axes:
            fig['layout'][a].update(
                type='linear',
                showline=False,
                showgrid=False,
                zeroline=False,
                mirror=False,
                fixedrange=False,
                showticklabels=False
            )

        (row_dendro_traces, col_dendro_traces) = self._sort_traces(
            dt['row'], dt['col'])

        for i in range(len(col_dendro_traces)):
            cdt = col_dendro_traces[i]
            cdt['name'] = ("Col Cluster %d" % i)
            cdt['line'] = dict(
                width=self._line_width[1]
            )
            cdt['hoverinfo'] = 'y+name'
            cluster_curve_numbers[len(fig.data)] = ['col', i]
            fig.append_trace(cdt, 1, 2)

        # row dendrogram (displays on left side)
        for i in range(len(row_dendro_traces)):
            rdt = row_dendro_traces[i]
            rdt['name'] = ("Row Cluster %d" % i)
            rdt['line'] = dict(
                width=self._line_width[0]
            )
            rdt['hoverinfo'] = 'x+name'
            cluster_curve_numbers[len(fig.data)] = ['row', i]
            fig.append_trace(rdt, 2, 1)

        # display row dendrogram sideways
        xaxis4 = fig['layout']['xaxis4']  # pylint: disable=invalid-sequence-index
        xaxis4.update(
            autorange='reversed'
        )

        # ensure that everything is aligned properly
        # with the heatmap
        yaxis4 = fig['layout']['yaxis4']  # pylint: disable=invalid-sequence-index
        yaxis4.update(
            scaleanchor='y5'
        )
        xaxis2 = fig['layout']['xaxis2']  # pylint: disable=invalid-sequence-index
        xaxis2.update(
            scaleanchor='x5'
        )

        if len(tickvals_col) == 0:
            tickvals_col = [10*i+5 for i in range(len(self._column_labels))]

        # add in all of the labels
        fig['layout']['xaxis5'].update(  # pylint: disable=invalid-sequence-index
            tickmode='array',
            tickvals=tickvals_col,
            ticktext=self._column_labels,
            tickfont=self._tick_font,
            showticklabels=True,
            side='bottom',
            showline=False,
            range=[min(tickvals_col)-5, max(tickvals_col)+5]
            # workaround for autoscale issues above; otherwise
            # the graph cuts off and must be scaled manually
        )

        if len(tickvals_row) == 0:
            tickvals_row = [10*i+5 for i in range(len(self._row_labels))]

        fig['layout']['yaxis5'].update(  # pylint: disable=invalid-sequence-index
            tickmode='array',
            tickvals=tickvals_row,
            ticktext=self._row_labels,
            tickfont=self._tick_font,
            showticklabels=True,
            side='right',
            showline=False,
            range=[min(tickvals_row), max(tickvals_row)]
        )

        # hide labels, if necessary
        for l in self._hide_labels:
            fig['layout'][l].update(
                ticks='',
                showticklabels=False
            )

        # recalculate the heatmap, if necessary
        if heatmap is None:

            # heatmap
            heat_data = self._data

            # symmetrize the heatmap about zero, if necessary
            if self._symmetric_value:
                heat_data = np.subtract(heat_data, np.mean(heat_data))

            heatmap = go.Heatmap(
                x=tickvals_col,
                y=tickvals_row,
                z=heat_data,
                colorscale=self._color_map,
                colorbar={
                    'xpad': 50
                }
            )

        fig.append_trace(heatmap, 2, 2)

        # hide all legends
        fig['layout'].update(
            showlegend=False,
        )

        # apply the display ratio
        row_ratio = 0
        col_ratio = 0

        # the argument can be either in list form or float form
        # first is ratio for row; second is ratio for column
        if self._display_ratio[0] != 0:
            row_ratio = \
                0.95/float(1 + int(1/self._display_ratio[0]))
        if self._display_ratio[1] != 0:
            col_ratio = \
                0.95/float(1 + int(1/self._display_ratio[1]))

        # the row/column labels take up 0.05 of the graph, and the rest
        # is taken up by the heatmap and dendrogram for each dimension

        # row: dendrogram, heatmap, row labels (left-to-right)
        # column: dendrogram, column labels, heatmap (top-to-bottom)

        # width adjustment for row dendrogram
        fig['layout']['xaxis1'].update(  # pylint: disable=invalid-sequence-index
            domain=[0, 0.95]
        )
        fig['layout']['xaxis2'].update(  # pylint: disable=invalid-sequence-index
            domain=[row_ratio, 0.95],
            anchor='y4'
        )
        fig['layout']['xaxis4'].update(  # pylint: disable=invalid-sequence-index
            domain=[0, row_ratio]
        )
        fig['layout']['xaxis5'].update(  # pylint: disable=invalid-sequence-index
            domain=[row_ratio, 0.95]
        )

        # height adjustment for column dendrogram
        fig['layout']['yaxis1'].update(  # pylint: disable=invalid-sequence-index
            domain=[1-col_ratio, 1]
        )
        fig['layout']['yaxis2'].update(  # pylint: disable=invalid-sequence-index
            domain=[1-col_ratio, 1]
        )
        fig['layout']['yaxis4'].update(  # pylint: disable=invalid-sequence-index
            domain=[0, 0.95-col_ratio]
        )
        fig['layout']['yaxis5'].update(  # pylint: disable=invalid-sequence-index
            domain=[0, 0.95-col_ratio]
        )

        fig['layout']['legend'] = dict(  # pylint: disable=unsupported-assignment-operation
            x=0.7,
            y=0.7
        )

        # annotations

        # axis settings for subplots that will display group labels
        axes = ['xaxis6', 'yaxis6',
                'xaxis8', 'yaxis8']

        for a in axes:
            fig['layout'][a].update(
                type='linear',
                showline=False,
                showgrid=False,
                zeroline=False,
                mirror=False,
                fixedrange=False,
                showticklabels=False,
            )

        # group labels for row dendrogram
        fig['layout']['yaxis6'].update(  # pylint: disable=invalid-sequence-index
            domain=[0, 0.95-col_ratio],
            scaleanchor='y5',
            scaleratio=1
        )
        if len(tickvals_row) > 0:
            fig['layout']['yaxis6'].update(  # pylint: disable=invalid-sequence-index
                range=[min(tickvals_row), max(tickvals_row)]
            )
        # padding between group label line and dendrogram
        fig['layout']['xaxis6'].update(  # pylint: disable=invalid-sequence-index
            domain=[0.95, 1],
            range=[-5, 1]
        )

        # group labels for column dendrogram
        fig['layout']['xaxis8'].update(  # pylint: disable=invalid-sequence-index
            domain=[row_ratio, 0.95],
            scaleanchor='x5',
            scaleratio=1
        )
        if len(tickvals_col) > 0:
            fig['layout']['xaxis8'].update(  # pylint: disable=invalid-sequence-index
                range=[min(tickvals_col), max(tickvals_col)]
            )
        fig['layout']['yaxis8'].update(  # pylint: disable=invalid-sequence-index
            domain=[0.95-col_ratio, 1-col_ratio],
            range=[-0.5, 0.5]
        )

        # get group label annotations and label traces
        row_group_labels, col_group_labels, row_annotations, col_annotations = \
            self._group_label_traces(
                row_dendro_traces,
                col_dendro_traces
            )
        # add annotations to graph
        fig['layout'].update(
            annotations=row_annotations + col_annotations
        )
        # add label traces to graph
        for rgl in row_group_labels:
            fig.append_trace(rgl, 2, 4)
        for cgl in col_group_labels:
            fig.append_trace(cgl, 4, 2)

        # set background colors
        fig['layout'].update(
            paper_bgcolor=self._paper_bg_color,
            plot_bgcolor=self._plot_bg_color
        )

        # finally add height and width
        fig['layout'].update(
            height=self._height,
            width=self._width
        )

        computed_traces = {'dendro_traces': dt,
                           'heatmap': heatmap,
                           'row_labels': self._row_labels,
                           'column_labels': self._column_labels}

        return (fig, computed_traces, cluster_curve_numbers)

    def _scale(
            self,
            dim
    ):
        """Returns standardized data based on user parameters.

        Keyword arguments:
        - dim (string): The dimension, row or column, to standardize across.

        Returns:
        - ndarray: An array containing the standardized data."""

        std = np.zeros(self._data.shape)

        if dim == 'row':
            std = scp.stats.zscore(self._data, axis=1)
        elif dim == 'column':
            std = scp.stats.zscore(self._data, axis=0)

        return std

    def _get_clusters(
            self
    ):
        """Clusters the data according to the specified dimensions.

        Returns:
        - tuple: The linkage matrices for the columns and/or rows."""

        Zcol = None
        Zrow = None

        # cluster along columns
        if self._cluster in ['col', 'all']:
            tmp = np.transpose(self._data)
            dcol = self._dist_fun(tmp, metric=self._col_dist)
            Zcol = self._link_fun(dcol, optimal_ordering=self._optimal_leaf_order)
        # cluster along rows only if 'all' is selected
        if self._cluster in ['row', 'all']:
            drow = self._dist_fun(self._data, metric=self._row_dist)
            Zrow = self._link_fun(drow, optimal_ordering=self._optimal_leaf_order)

        return (Zcol, Zrow)

    def _compute_clustered_data(
            self
    ):
        """Get the traces that need to be plotted for the row and column
        dendrograms, and update the ordering of the 2D data array,
        row labels, and column labels to match the reordered
        dendrogram leaves.

        Returns:
        - dict: A dictionary containing entries for the row and column
        dendrogram traces.
        - ndarray: The original 2D data array that has been reordered to
        match the ordering of the row and column dendrogram leaves.
        - list: A list of the row labels that have been reordered to match
        the ordering of the row dendrogram leaves.
        - list: a list of the column labels that have been reordered to match
        the ordering of the column dendrogram leaves.

        """

        # initialize return dict
        trace_list = {
            'col': [],
            'row': []
        }

        # first, compute the clusters
        (Zcol, Zrow) = self._get_clusters()

        clustered_column_labels = self._column_labels
        clustered_row_labels = self._row_labels

        # calculate dendrogram from clusters; sch.dendrogram returns sets
        # of four coordinates that make up the 'u' shapes in the dendrogram
        if Zcol is not None:
            Pcol = sch.dendrogram(Zcol, orientation='top',
                                  color_threshold=self._color_threshold['col'],
                                  labels=self._column_labels, no_plot=True)
            clustered_column_labels = scp.array(Pcol['ivl'])
            trace_list['col'] = self._color_dendro_clusters(Pcol, 'col')

        if Zrow is not None:
            Prow = sch.dendrogram(Zrow, orientation='left',
                                  color_threshold=self._color_threshold['row'],
                                  labels=self._row_labels, no_plot=True)
            # need to flip the coordinates for the row dendrogram
            Prow_tmp = {
                'icoord': Prow['dcoord'],
                'dcoord': Prow['icoord'],
                'color_list': Prow['color_list']
            }
            clustered_row_labels = scp.array(Prow['ivl'])
            trace_list['row'] = self._color_dendro_clusters(Prow_tmp, 'row')

        # now, we need to rearrange the data array to fit the labels

        # first find the order in which to shuffle the data
        rl_indices = [list(clustered_row_labels).index(r)
                      for r in list(self._row_labels)]
        cl_indices = [list(clustered_column_labels).index(c)
                      for c in list(self._column_labels)]

        # modify the data here; first shuffle rows,
        # then transpose and shuffle columns,
        # then transpose again
        clustered_data = self._data[rl_indices].T[cl_indices].T

        return trace_list, clustered_data, clustered_row_labels, clustered_column_labels

    def _color_dendro_clusters(
            self,
            P,
            dim
    ):
        """Colors each cluster below the color threshold separately.

        Keyword arguments:
        - P (dict): The x and y values of the dendrogram traces, along
            with the list of trace colors returned by sch.dendrogram
        - dim (string): The dimension of the clusters.

        Returns:
        - list: The list of colored traces for the dendrogram."""

        traces = []

        icoord = scp.array(P['icoord'])
        dcoord = scp.array(P['dcoord'])

        color_list = self._cluster_colors(P['color_list'], dim)

        # dict w/ keys being the color code and values being another dict
        # specifying icoords and dcoords for that cluster
        clusters = {}

        for i in range(len(color_list)):
            # the current trace
            c = color_list[i]
            clusters[str(c['cluster'])] = {
                'color': c['color'],
                'icoords': [icoord[j] for j in range(len(icoord))
                            if color_list[j]['cluster'] == c['cluster']],
                'dcoords': [dcoord[j] for j in range(len(dcoord))
                            if color_list[j]['cluster'] == c['cluster']]
            }

        for c in clusters:

            # all of the coordinates
            icoords = clusters[c]['icoords']
            dcoords = clusters[c]['dcoords']

            # initialize all x and y values for this cluster trace
            x = np.array([])
            y = np.array([])

            for j in range(len(icoords)):
                x = np.append(x, icoords[j])
                y = np.append(y, dcoords[j])

                # append nan to prevent links
                x = np.append(x, np.nan)
                y = np.append(y, np.nan)

            traces.append(dict(
                x=x,
                y=y,
                type='scatter',
                mode='lines',
                marker=dict(color=clusters[c]['color'])
            ))

        return traces

    def _cluster_colors(
            self,
            clist,
            dim
    ):
        """Returns a set of n unique colours for each cluster in the dendrogram.

        Keyword arguments:
        - clist (list): The color list returned by dendrogram.
        - dim (string): The dimension of the clusters to color.

        Returns:
        list: A list of RGB strings."""
        # the colors repeat; get how many repetitions there are

        # the colors go through cycles of g, r, c, m, y, k
        cycles = []
        # store a string representing the current cycle
        curr_cycle = ''

        # iterate through the list of colors
        i = 0
        while i < len(clist):
            # add the color to the current cycle
            curr_cycle += clist[i]
            # treat the end of the list as the end of a cycle
            if i == len(clist)-1:
                cycles.append(curr_cycle)
                break
            # otherwise, the end of a cycle is signified by
            # a sequence k, g - however, we also have b for
            # the links above the color threshold; so we
            # include this as well
            if clist[i] in ['k', 'b'] and clist[i+1] == 'g':
                cycles.append(curr_cycle)
                curr_cycle = ''
            # finally, increment the counter
            i += 1

        color_list = []
        bg_color = 'rgb(0,0,0)'

        # each element in 'cycles' contains a full cycle of 6 colors
        # (at most)
        # so, we need 6 times the number of cycles
        n = 6*len(cycles)

        # fill in the user-provided color list if possible
        if self._color_list is not None and dim in self._color_list:
            color_list = self._color_list[dim]
            # if there aren't enough colors, repeat the list
            if len(color_list) < n and len(color_list) > 0:
                color_list = color_list * (int(n/len(color_list)) + 1)

        else:
            # first, get the number of divisions
            # take cube root of n, since we will need to make
            # at least n^3 unique colors (filling 3 spots)
            base = int(np.cbrt(n)) + 1
            # rgb values to use
            vals = np.linspace(0, 255, base).astype(int)

            # pure rgb, including black
            pure = []
            # mixes of 2 pure rgb values
            mixed = []
            # other colors
            other = []

            for i in range(np.power(base, 3) - 1):

                # the color represented in n-ary
                c = str(np.base_repr(i, base=base, padding=3))[-3:]

                r = int(vals[int(c[0])])
                g = int(vals[int(c[1])])
                b = int(vals[int(c[2])])

                # color, represented as a string
                color = "rgb(%d,%d,%d)" % (r, g, b)

                # priority is all of the pure colors
                if (r == 255 or g == 255 or b == 255) and r + g + b == 255:
                    pure.append(color)
                    # then, all of the mixtures of pure colors
                elif (r == 0 or g == 0 or b == 0) and r + g + b == 510:
                    mixed.append(color)
                    # all of the intermediate colors that are created
                else:
                    other.append(color)

            # avoid similarity of colors
            shuffle(other)

            color_list = pure + mixed + other

        # this will be returned
        colors = []

        # get the color for the background trace, if one is supplied
        if self._color_list is not None and 'bg' in self._color_list:
            bg_color = self._color_list['bg']

        # the sequence
        seq = ['g', 'r', 'c', 'm', 'y', 'k']

        for i in range(len(cycles)):
            tmp = []
            for s in seq:
                for j in range(cycles[i].count(s)):
                    tmp.append({'color': color_list[i*6 + seq.index(s)],
                                'cluster': i*6 + seq.index(s)})
            # get all indices of 'b', the links above threshold
            bs = [j for j in range(len(cycles[i])) if cycles[i][j] == 'b']
            for index in bs:
                # may need to change this color
                # depending on which are generated
                tmp.insert(index,
                           {'color': bg_color,
                            'cluster': -1})

            colors = colors + tmp

        return colors

    def _sort_traces(
            self,
            rdt,
            cdt
    ):
        """Sorts row dendrogram clusters and column dendrogram clusters
        so that the background trace (above threshold) is trace 0
        and all other traces are ordered top-to-bottom (row dendrogram)
        or left-to-right (column dendrogram).

        Keyword arguments:
        - rdt (list[dict]): The row dendrogram cluster traces.
        - cdt (list[dict]): The column dendrogram cluster traces.

        Returns:
        - tuple: The sorted row dendrogram clusters and column
                        dendrogram clusters."""

        tmp_rdt = []
        tmp_cdt = []

        if len(rdt) > 0:
            # first, find background trace: (max 'x')
            rdt.sort(key=lambda t: -1*max(list(t['x'])))
            tmp_rdt.append(rdt[0])
            # then, sort top-to-bottom
            r = rdt[1:]
            r.sort(key=lambda t: -1*min(list(t['y'])))
            tmp_rdt += r
        if len(cdt) > 0:
            # background trace has max 'y'
            cdt.sort(key=lambda t: -1*max(list(t['y'])))
            tmp_cdt.append(cdt[0])
            # sort left to right
            c = cdt[1:]
            c.sort(key=lambda t: min(list(t['x'])))
            tmp_cdt += c

        return(tmp_rdt, tmp_cdt)

    def _group_label_traces(
            self,
            row_clusters,
            col_clusters
    ):
        """Calculates the traces and annotations that correspond to group
        labels.

        Keyword arguments:
        - row_clusters (list[dict]): List of all row traces (each
            trace corresponds to a cluster)
        - col_clusters (list[dict]): List of all column traces (each
            trace corresponds to a cluster)

        Returns:
        - tuple: The row label traces, column label traces, row group
            annotations, and column group annotations."""

        row_group_labels = []
        col_group_labels = []

        row_annotations = []
        col_annotations = []

        for rgm in self._row_group_marker:
            if len(row_clusters) == 0:
                break
            if rgm['group'] >= len(row_clusters):
                continue
            # get upper and lower bounds of group
            ymin = min(row_clusters[rgm['group']]['y'])
            ymax = max(row_clusters[rgm['group']]['y'])
            trace = go.Scatter(
                x=[0, 0],
                y=[ymin, ymax],
                mode='lines',
                line=dict(
                    width=6,
                    color=rgm['color']
                ),
                marker=dict(
                    size=0
                ),
                hoverinfo='none'
            )
            row_group_labels.append(trace)
            row_annotations.append(dict(
                x=0.5, y=1/2*(ymin + ymax),
                xref='x6', yref='y6',
                text=rgm['annotation'],
                font=self._annotation_font,
                showarrow=False,
                xanchor='left'
            ))

        for cgm in self._col_group_marker:
            if len(col_clusters) == 0:
                break
            if cgm['group'] >= len(col_clusters):
                continue
            # get leftmost and rightmost bounds of group
            xmin = min(col_clusters[cgm['group']]['x'])
            xmax = max(col_clusters[cgm['group']]['x'])
            trace = go.Scatter(
                x=[xmin, xmax],
                y=[0, 0],
                mode='lines',
                line=dict(
                    width=6,
                    color=cgm['color']
                ),
                marker=dict(
                    size=0
                ),
                hoverinfo='none'
            )
            col_group_labels.append(trace)
            col_annotations.append(dict(
                x=1/2*(xmin + xmax), y=-0.5,
                xref='x8', yref='y8',
                text=cgm['annotation'],
                font=self._annotation_font,
                showarrow=False
            ))

        return (row_group_labels, col_group_labels, row_annotations, col_annotations)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Ideogram(Component):
    """A Ideogram component.
The Ideogram component is used to draw and animate genome-wide
datasets for organisms such as human, mouse, and any other
eukaryote. The Ideogram component can be used to compare
homologous features between chromosomes, and depict
haploid, diploid, aneuploid genomes. It can also display
annotations on genomic data using histograms and overlays.

Reference: https://eweitz.github.io/ideogram/
Component's props: https://github.com/eweitz/ideogram/blob/master/api.md

Keyword arguments:
- id (string; required): The ID used to identify this component in Dash callbacks and used to identify Ideogram
instances.
- style (dict; optional): The component's inline styles
- className (string; optional): The CSS class of the component wrapper
- annotationsLayout (a value equal to: 'tracks', 'histogram', 'overlay'; optional): Layout of ideogram annotations.
One of "tracks", "histogram", or "overlay".

"tracks": display annotations in tracks beside each chromosome.

"histogram": display annotations in a histogram. Clusters annotations by location. Each
cluster/bin is shown as a bar, the height of which represents the number of annotations on
genomic range.

"overlay": display annotations directly over chromosomes.
- annotations (list; optional): A list of annotation objects. Annotation objects can also have a name, color, shape, and
track index. At the moment there is more keys specified and the docs need updating.
- annotationsPath (string; optional): An absolute or relative URL directing to a JSON file containing annotation objects (JSON).
- annotationsData (string; optional): Use this prop in a dash callback to return annotationData when hovered.
It is read-only, i.e., it cannot be used with dash.dependencies.Output but only with
dash.dependencies.Input
- annotationTracks (list; optional): A list of objects with metadata for each track, e.g., id, display name, color, shape.
- annotationHeight (number; optional): Not used if annotationsLayout is set to "overlay".
The height of histogram bars or the size of annotations tracks symbols
- annotationsColor (string; optional): Color of annotations.
- histogramScaling (a value equal to: 'absolute', 'relative'; optional): Scaling of histogram bars height
Only used if annotationsLayout is set to "histogram".
One of "absolute" or "relative".

"absolute": sets bar height relative to tallest bar in all chromosomes.
"relative": sets bar height relative to tallest bar in each chromosome.
- barWidth (number; optional): Pixel width of histogram bars.
Only used if annotationsLayout is set to "histogram".
- showAnnotTooltip (boolean; optional): Whether to show a tooltip upon mousing over an annotation.
- assembly (string; optional): Default: latest RefSeq assembly for specified organism.
The genome assembly to display.
Takes assembly name (e.g., "GRCh37"),
RefSeq accession (e.g., "GCF_000306695.2"),
or GenBank accession (e.g., "GCA_000005005.5")
- brush (string; optional): Genomic coordinate range (e.g., "chr1:104325484-119977655") for a brush on a
chromosome. Useful when ideogram consists of one chromosome and you want to be
able to focus on a region within that chromosome,
and create an interactive sliding window to other regions
- brushData (optional): A dash callback that is activated when the 'brush' prop is used in component.
It will return an dictionary like so:

{'start': <value>, 'end': <value>, 'extent': <value>}

where start is the left most edge, end is right most edge, and extent is the total width of
the brush.
It is read-only, i.e., it cannot be used with dash.dependencies.Output but only with
dash.dependencies.Input. brushData has the following type: dict containing keys 'start', 'end', 'extent'.
Those keys have the following types:
  - start (string; optional)
  - end (string; optional)
  - extent (string; optional)
- container (string; optional): CSS styling and the id of the container holding the Ideogram in
react-ideogram.js, this is where all the d3 magic happens.
- chrHeight (number; optional): The pixel height of the tallest chromosome in the ideogram
- chrMargin (number; optional): The pixel space of margin between each chromosome.
- chrWidth (number; optional): The pixel width of each chromosome.
- chromosomes (list | dict; optional): A list of the names of chromosomes to display. Useful for depicting a subset of the
chromosomes in the genome, e.g., a single chromosome.

If Homology (between two different species):
Ex: chromosomes={
      'human': ['1'],
      'mouse': ['4']
}

General case to specify specific chromosomes:
Ex: chromosomes=['1', '2']
- dataDir (string; optional): Absolute or relative URL of the directory containing data needed to draw banded chromosomes.
You will need to set up your own database to grab data from a custom database.
- organism (string | number; optional): Organism(s) to show chromosomes for. Supply organism's name as a string (e.g., "human") or
organism's NCBI Taxonomy ID (taxid, e.g., 9606) to display chromosomes from a single
organism, or an array of organisms' names or taxids to display chromosomes from multiple
species.
- localOrganism (dict; optional): Provide local JSON organism into this prop from a local user JSON file.
DataDir must not be initialized.
- homology (optional): Used to compare two chromosomes with each other.
The keys "chrOne" and "chrTwo" represent one chromosome each. Organism is the taxID or name.
Start is an array, containing start one and start two, in this order. Stop is an array,
containing stop one, and stop two, in this order.
Ex: homology={
    "chrOne": {
        organism": "9606",
        "start": [50000, 155701383],
        "stop": [900000, 156030895]
    },
    "chrTwo": {
        organism": "10090",
        "start": [10001, 50000000],
        "stop": [2781479, 57217415]
    }
}. homology has the following type: dict containing keys 'chrOne', 'chrTwo'.
Those keys have the following types:
  - chrOne (optional): . chrOne has the following type: dict containing keys 'organism', 'start', 'stop'.
Those keys have the following types:
  - organism (string; required)
  - start (list; optional)
  - stop (list; optional)
  - chrTwo (optional): . chrTwo has the following type: dict containing keys 'organism', 'start', 'stop'.
Those keys have the following types:
  - organism (string; required)
  - start (list; optional)
  - stop (list; optional)
- perspective (a value equal to: 'comparative'; optional): Use perspective: 'comparative' to enable annotations between two chromosomes,
either within the same organism or different organisms. Used for homology.
- fullChromosomeLabels (boolean; optional): Whether to include abbreviation species name in chromosome label. Used for homology.
- resolution (number; optional): The resolution of cytogenetic bands to show for each chromosome.
The quantity refers to an approximate value in bands per haploid set (bphs).
One of 450, 550, or 850.
- filterable (boolean; optional): Whether annotations should be filterable or not.
- orientation (a value equal to: 'vertical', 'horizontal'; optional): The orientation of chromosomes on the page.
- ploidy (number; optional): The ploidy - number of chromosomes to depict for each chromosome set.
- ploidyDesc (list; optional): Description of ploidy in each chromosome set in terms of ancestry composition.
- ancestors (dict; optional): A map associating ancestor labels to colors. Used to color
chromosomes from different ancestors in polyploid genomes.
- rangeSet (list; optional): List of objects describing segments of recombination among chromosomes in a chromosome set.
- rotatable (boolean; optional): Whether chromosomes are rotatable on click.
- rotated (boolean; optional): Dash callback that returns true if rotated, and false if not.
- sex (a value equal to: 'male', 'female'; optional): Useful for omitting chromosome Y in female animals.
Currently only supported for organisms that use XY sex-determination.
- showChromosomeLabels (boolean; optional): Whether to show chromosome labels, e.g., 1, 2, 3, X, Y.
- showBandLabels (boolean; optional): Whether to show cytogenetic band labels, e.g., 1q21
- showFullyBanded (boolean; optional): Whether to show fully banded chromosomes for genomes that have sufficient data. Useful for
showing simpler chromosomes of cytogenetically well-characterized organisms, e.g., human,
beside chromosomes of less studied organisms, e.g., chimpanzee.
- showNonNuclearChromosomes (boolean; optional): Whether to show non-nuclear chromosomes,
e.g., for mitochondrial (MT) and chloroplast (CP) DNA."""
    @_explicitize_args
    def __init__(self, id=Component.REQUIRED, style=Component.UNDEFINED, className=Component.UNDEFINED, annotationsLayout=Component.UNDEFINED, annotations=Component.UNDEFINED, annotationsPath=Component.UNDEFINED, annotationsData=Component.UNDEFINED, annotationTracks=Component.UNDEFINED, annotationHeight=Component.UNDEFINED, annotationsColor=Component.UNDEFINED, histogramScaling=Component.UNDEFINED, barWidth=Component.UNDEFINED, showAnnotTooltip=Component.UNDEFINED, assembly=Component.UNDEFINED, brush=Component.UNDEFINED, brushData=Component.UNDEFINED, container=Component.UNDEFINED, chrHeight=Component.UNDEFINED, chrMargin=Component.UNDEFINED, chrWidth=Component.UNDEFINED, chromosomes=Component.UNDEFINED, dataDir=Component.UNDEFINED, organism=Component.UNDEFINED, localOrganism=Component.UNDEFINED, homology=Component.UNDEFINED, perspective=Component.UNDEFINED, fullChromosomeLabels=Component.UNDEFINED, resolution=Component.UNDEFINED, filterable=Component.UNDEFINED, orientation=Component.UNDEFINED, ploidy=Component.UNDEFINED, ploidyDesc=Component.UNDEFINED, ancestors=Component.UNDEFINED, rangeSet=Component.UNDEFINED, rotatable=Component.UNDEFINED, rotated=Component.UNDEFINED, sex=Component.UNDEFINED, showChromosomeLabels=Component.UNDEFINED, showBandLabels=Component.UNDEFINED, showFullyBanded=Component.UNDEFINED, showNonNuclearChromosomes=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'style', 'className', 'annotationsLayout', 'annotations', 'annotationsPath', 'annotationsData', 'annotationTracks', 'annotationHeight', 'annotationsColor', 'histogramScaling', 'barWidth', 'showAnnotTooltip', 'assembly', 'brush', 'brushData', 'container', 'chrHeight', 'chrMargin', 'chrWidth', 'chromosomes', 'dataDir', 'organism', 'localOrganism', 'homology', 'perspective', 'fullChromosomeLabels', 'resolution', 'filterable', 'orientation', 'ploidy', 'ploidyDesc', 'ancestors', 'rangeSet', 'rotatable', 'rotated', 'sex', 'showChromosomeLabels', 'showBandLabels', 'showFullyBanded', 'showNonNuclearChromosomes']
        self._type = 'Ideogram'
        self._namespace = 'dash_bio'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'style', 'className', 'annotationsLayout', 'annotations', 'annotationsPath', 'annotationsData', 'annotationTracks', 'annotationHeight', 'annotationsColor', 'histogramScaling', 'barWidth', 'showAnnotTooltip', 'assembly', 'brush', 'brushData', 'container', 'chrHeight', 'chrMargin', 'chrWidth', 'chromosomes', 'dataDir', 'organism', 'localOrganism', 'homology', 'perspective', 'fullChromosomeLabels', 'resolution', 'filterable', 'orientation', 'ploidy', 'ploidyDesc', 'ancestors', 'rangeSet', 'rotatable', 'rotated', 'sex', 'showChromosomeLabels', 'showBandLabels', 'showFullyBanded', 'showNonNuclearChromosomes']
        self.available_wildcard_properties =            []

        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        for k in ['id']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Ideogram, self).__init__(**args)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Molecule3dViewer(Component):
    """A Molecule3dViewer component.
The Molecule3dViewer component is used to render schematic diagrams
of biomolecules. It can display ribbon-structure diagrams, or
render atoms in the molecule as sticks or spheres.
Read more about the component here:
https://github.com/Autodesk/molecule-3d-for-react

Keyword arguments:
- id (string; optional): The ID used to identify this component in callbacks
- selectionType (a value equal to: 'atom', 'residue', 'chain'; optional): The selection type - may be atom, residue or chain
- backgroundColor (string; optional): Property to change the background color of the molecule viewer
- backgroundOpacity (number; optional): Property to change the backgroun opacity - ranges from 0 to 1
- styles (list; optional): Property that can be used to change the representation of
the molecule. Options include sticks, cartoon and sphere
- modelData (optional): The data that will be used to display the molecule in 3D
The data will be in JSON format
and should have two main dictionaries - atoms, bonds. modelData has the following type: dict containing keys 'atoms', 'bonds'.
Those keys have the following types:
  - atoms (list; optional)
  - bonds (list; optional)
- atomLabelsShown (boolean; optional): Property to either show or hide labels
- selectedAtomIds (list; optional): Property that stores a list of all selected atoms
- labels (list; optional): labels corresponding to the atoms of the molecule"""
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, selectionType=Component.UNDEFINED, backgroundColor=Component.UNDEFINED, backgroundOpacity=Component.UNDEFINED, styles=Component.UNDEFINED, modelData=Component.UNDEFINED, atomLabelsShown=Component.UNDEFINED, selectedAtomIds=Component.UNDEFINED, labels=Component.UNDEFINED, onRenderNewData=Component.UNDEFINED, onChangeSelection=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'selectionType', 'backgroundColor', 'backgroundOpacity', 'styles', 'modelData', 'atomLabelsShown', 'selectedAtomIds', 'labels']
        self._type = 'Molecule3dViewer'
        self._namespace = 'dash_bio'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'selectionType', 'backgroundColor', 'backgroundOpacity', 'styles', 'modelData', 'atomLabelsShown', 'selectedAtomIds', 'labels']
        self.available_wildcard_properties =            []

        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        for k in []:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Molecule3dViewer, self).__init__(**args)



from __future__ import absolute_import

import pandas as pd
from pandas.api.types import is_numeric_dtype
import numpy as np

import plotly.graph_objs as go

SUGGESTIVE_LINE_LABEL = "suggestive line"
GENOMEWIDE_LINE_LABEL = "genomewide line"


def _get_hover_text(df, snpname=None, genename=None, annotationname=None):
    """format the hover text used for Manhattan and Volcano Plots
    :param (dataFrame): A pandas dataframe
    :param (string) snpname: A string denoting the column name for the SNP
    names (e.g. rs number). More generally, this column could be anything
    that identifies each point being plotted. For example,
    in an Epigenomewide association study (EWAS) this could be the probe
    name or cg number. This column should be a character. This argument is
    optional, however it is necessary to specify if you want to
    highlight points on the plot using the highlight argument in the
    figure method.
        Default = None
    :param (string) genename: A string denoting the column name for the
    GENE names.
    More generally this could be any annotation information that you
    want to include in the plot.
        Default = None
    :param (string) annotationname: A string denoting the column name
    for an annotation. This could be any annotation information that you
    want to include in the plot (e.g. zscore, effect size, minor allele
    frequency).
        Default = None
    """
    hover_text = ""
    if snpname is not None and snpname in df.columns:
        hover_text = "SNP: " + df[snpname].astype(str)

    if genename is not None and genename in df.columns:
        hover_text = hover_text \
                     + "<br>GENE: " \
                     + df[genename].astype(str)

    if annotationname is not None and annotationname in df.columns:
        hover_text = hover_text \
                     + "<br>" \
                     + df[annotationname].astype(str)
    return hover_text


def ManhattanPlot(
        dataframe,
        chrm="CHR",
        bp="BP",
        p="P",
        snp="SNP",
        gene="GENE",
        annotation=None,
        logp=True,
        title="Manhattan Plot",
        showgrid=True,
        xlabel=None,
        ylabel='-log10(p)',
        point_size=5,
        showlegend=True,
        col=None,
        suggestiveline_value=-np.log10(1e-8),
        suggestiveline_color='#636efa',
        suggestiveline_width=1,
        genomewideline_value=-np.log10(5e-8),
        genomewideline_color='#EF553B',
        genomewideline_width=1,
        highlight=True,
        highlight_color="red",
):
    """Returns a figure for a manhattan plot.

Keyword arguments:
- dataframe (dataframe; required): A pandas dataframe which must contain at
    least the following  three columns:
            - the chromosome number
            - genomic base-pair position
            - a numeric quantity to plot such as a p-value or zscore
- chrm (string; optional): A string denoting the column name for the
    chromosome.  This column must be float or integer.  Minimum number
    of chromosomes required is 1. If you have X, Y, or MT chromosomes,
    be sure to renumber these 23, 24, 25, etc. (Default: "CHR")
- bp (string; optional): A string denoting the column name for the
    chromosomal position. (Default: "BP")
- p (string; optional): A string denoting the column name for the
    float quantity to be plotted on the y-axis. This column must be
    numeric. This does not have to be a p-value. It can be any
    numeric quantity such as peak heights, bayes factors, test
    statistics. If it is not a p-value, make sure to set logp = FALSE.
    (Default: "P")
- snp (string; optional): A string denoting the column name for the
    SNP names (e.g. rs number). More generally, this column could be
    anything that identifies each point being plotted. For example, in
    an Epigenomewide association study (EWAS) this could be the probe
    name or cg number. This column should be a character. This
    argument is optional, however it is necessary to specify if you
    want to highlight points on the plot using the highlight argument
    in the figure method. (Default: "SNP")
- gene (string; optional): A string denoting the column name for the
    GENE names. This column could be a string or a float. More
    generally this could be any annotation information that you want
    to include in the plot. (Default: "GENE")
- annotation (string; optional): A string denoting the column name for
    an annotation. This column could be a string or a float.  This
    could be any annotation information that you want to include in
    the plot (e.g. zscore, effect size, minor allele frequency).
    (Default: None)
- logp (bool; optional): If True, the -log10 of the p-value is
    plotted.  It isn't very useful to plot raw p-values; however,
    plotting the raw value could be useful for other genome-wide plots
    (e.g., peak heights, bayes factors, test statistics, other
    "scores", etc.) (Default: True)
- title (string; optional) The title of the graph. (Default: "Manhattan Plot")
- showgrid (bool; optional): Boolean indicating whether gridlines should be
    shown. (Default: True)
- xlabel (string; optional): Label of the x axis. (Default: None)
- ylabel: (string; optional): Label of the y axis. (Default:
    "-log10(p)")
- point_size (number; optional): Size of the points of the Scatter
    plot. (Default: 5)
- showlegend (bool; optional): Boolean indicating whether legends should be
    shown. (Default: True)
- col (string; optional): A string representing the color of the
    points of the Scatter plot. Can be in any color format accepted by
    plotly_js graph_objs. (Default: None) Default = None
- suggestiveline_value (bool/float; optional): A value which must
    be False to deactivate the option, or a numerical value
    corresponding to the p-value at which the line should be drawn.
    The line has no influence on the data points. (Default:
    -np.log10(1e-8))
- suggestiveline_color (string; optional): Color of the suggestive
  line. (Default: "grey")
- suggestiveline_width (number): Width of the suggestive
        line. (Default: 2)
- genomewideline_value (bool/float; optional): A boolean which must be
    False to deactivate the option, or a numerical value corresponding
    to the p-value above which the data points are considered
    significant. (Default: -np.log10(5e-8))
- genomewideline_color (string; optional): Color of the genome wide
    line. Can be in any color format accepted by plotly_js
    graph_objs. (Default: "red")
- genomewideline_width (number; optional): Width of the genome wide
  line. (Default: 1)
- highlight (bool; optional): turning on/off the highlighting of data points
    considered significant. (Default: True)
- highlight_color (string; optional): Color of the data points
    highlighted because they are significant Can be in any color
    format accepted by plotly_js graph_objs. (Default: "red")

    # ...
    Example 1: Random Manhattan Plot
    '''
    dataframe = pd.DataFrame(
        np.random.randint(0,100,size=(100, 3)),
        columns=['P', 'CHR', 'BP'])
    fig = create_manhattan(dataframe, title='XYZ Manhattan plot')

    plotly.offline.plot(fig, image='png')
    '''

    """

    mh = _ManhattanPlot(
        dataframe,
        chrm=chrm,
        bp=bp,
        p=p,
        snp=snp,
        gene=gene,
        annotation=annotation,
        logp=logp
    )

    return mh.figure(
        title=title,
        showgrid=showgrid,
        xlabel=xlabel,
        ylabel=ylabel,
        point_size=point_size,
        showlegend=showlegend,
        col=col,
        suggestiveline_value=suggestiveline_value,
        suggestiveline_color=suggestiveline_color,
        suggestiveline_width=suggestiveline_width,
        genomewideline_value=genomewideline_value,
        genomewideline_color=genomewideline_color,
        genomewideline_width=genomewideline_width,
        highlight=highlight,
        highlight_color=highlight_color
    )


class _ManhattanPlot():

    def __init__(
            self,
            x,
            chrm="CHR",
            bp="BP",
            p="P",
            snp="SNP",
            gene="GENE",
            annotation=None,
            logp=True
    ):
        """
        Keyword arguments:
        - dataframe (dataframe; required): A pandas dataframe which
        must contain at least the following three columns:
            - the chromosome number
            - genomic base-pair position
            - a numeric quantity to plot such as a p-value or zscore
        - chrm (string; optional): A string denoting the column name for the
        chromosome.  This column must be float or integer.  Minimum number
        of chromosomes required is 1. If you have X, Y, or MT chromosomes,
        be sure to renumber these 23, 24, 25, etc. (Default: "CHR")
        - bp (string; optional): A string denoting the column name for the
        chromosomal position. (Default: "BP")
        - p (string; optional): A string denoting the column name for the
        float quantity to be plotted on the y-axis. This column must be
        numeric. This does not have to be a p-value. It can be any
        numeric quantity such as peak heights, bayes factors, test
        statistics. If it is not a p-value, make sure to set logp = FALSE.
        (Default: "P")
        - snp (string; optional): A string denoting the column name for the
        SNP names (e.g. rs number). More generally, this column could be
        anything that identifies each point being plotted. For example, in
        an Epigenomewide association study (EWAS) this could be the probe
        name or cg number. This column should be a character. This
        argument is optional, however it is necessary to specify if you
        want to highlight points on the plot using the highlight argument
        in the figure method. (Default: "SNP")
        - gene (string; optional): A string denoting the column name for the
        GENE names. This column could be a string or a float. More
        generally this could be any annotation information that you want
        to include in the plot. (Default: "GENE")
        - annotation (string; optional): A string denoting the column name for
        an annotation. This column could be a string or a float.  This
        could be any annotation information that you want to include in
        the plot (e.g. zscore, effect size, minor allele frequency).
        (Default: None)
        - logp (bool; optional): If True, the -log10 of the p-value is
        plotted.  It isn't very useful to plot raw p-values; however,
        plotting the raw value could be useful for other genome-wide plots
        (e.g., peak heights, bayes factors, test statistics, other
        "scores", etc.) (Default: True)

        Returns:
        - A ManhattanPlot object."""

        # checking the validity of the arguments

        # Make sure you have chrm, bp and p columns and that they are of
        # numeric type
        if chrm not in x.columns.values:
            raise KeyError("Column %s not found in 'x' data.frame" % chrm)
        else:
            if not is_numeric_dtype(x[chrm].dtype):
                raise TypeError("%s column should be numeric. Do you have "
                                "'X', 'Y', 'MT', etc? If so change to "
                                "numbers and try again." % chrm)

        if bp not in x.columns.values:
            raise KeyError("Column %s not found in 'x' data.frame" % bp)
        else:
            if not is_numeric_dtype(x[bp].dtype):
                raise TypeError("%s column should be numeric type" % bp)

        if p not in x.columns.values:
            raise KeyError("Column %s not found in 'x' data.frame" % p)
        else:
            if not is_numeric_dtype(x[p].dtype):
                raise TypeError("%s column should be numeric type" % p)

        # Create a new DataFrame with columns named after chrm, bp, and p.
        self.data = pd.DataFrame(data=x[[chrm, bp, p]])

        if snp is not None:
            if snp not in x.columns.values:
                # Warn if you don't have a snp column
                raise KeyError(
                    "snp argument specified as %s but column not found in "
                    "'x' data.frame" % snp)
            else:
                # If the input DataFrame has a snp column, add it to the new
                # DataFrame
                self.data[snp] = x[snp]

        if gene is not None:
            if gene not in x.columns.values:
                # Warn if you don't have a gene column
                raise KeyError(
                    "gene argument specified as %s but column not found in "
                    "'x' data.frame" % gene)
            else:
                # If the input DataFrame has a gene column, add it to the new
                # DataFrame
                self.data[gene] = x[gene]

        if annotation is not None:
            if annotation not in x.columns.values:
                # Warn if you don't have an annotation column
                raise KeyError(
                    "annotation argument specified as %s but column not "
                    "found in 'x' data.frame" % annotation
                )
            else:
                # If the input DataFrame has a gene column, add it to the new
                # DataFrame
                self.data[annotation] = x[annotation]

        self.xlabel = ""
        self.ticks = []
        self.ticksLabels = []
        self.nChr = len(x[chrm].unique())
        self.chrName = chrm
        self.pName = p
        self.snpName = snp
        self.geneName = gene
        self.annotationName = annotation
        self.logp = logp

        # Set positions, ticks, and labels for plotting

        self.index = 'INDEX'
        self.pos = 'POSITION'

        # Fixes the bug where one chromosome is missing by adding a sequential
        # index column.
        idx = 0
        for i in self.data[chrm].unique():
            idx = idx + 1
            self.data.loc[self.data[chrm] == i, self.index] = int(idx)
        # Set the type to be the same as provided for chrm column
        self.data[self.index] = \
            self.data[self.index].astype(self.data[chrm].dtype)

        # This section sets up positions and ticks. Ticks should be placed in
        # the middle of a chromosome. The new pos column is added that keeps
        # a running sum of the positions of each successive chromosome.
        # For example:
        # chrm bp pos
        # 1   1  1
        # 1   2  2
        # 2   1  3
        # 2   2  4
        # 3   1  5

        if self.nChr == 1:
            # For a single chromosome
            self.data[self.pos] = self.data[bp]
            self.ticks.append(int(len(self.data[self.pos]) / 2.) + 1)
            self.xlabel = "Chromosome %s position" % (self.data[chrm].unique())
            self.ticksLabels = self.ticks
        else:
            # For multiple chromosomes
            lastbase = 0
            for i in self.data[self.index].unique():
                if i == 1:
                    self.data.loc[self.data[self.index] == i, self.pos] = \
                        self.data.loc[self.data[self.index] == i, bp].values
                else:
                    prevbp = self.data.loc[self.data[self.index] == i - 1, bp]
                    # Shift the basepair position by the largest bp of the
                    # current chromosome
                    lastbase = lastbase + prevbp.iat[-1]

                    self.data.loc[self.data[self.index] == i, self.pos] = \
                        self.data.loc[self.data[self.index] == i, bp].values \
                        + lastbase

                tmin = min(self.data.loc[self.data[self.index] == i, self.pos])
                tmax = max(self.data.loc[self.data[self.index] == i, self.pos])
                self.ticks.append(int((tmin + tmax) / 2.) + 1)

            self.xlabel = 'Chromosome'
            self.data[self.pos] = self.data[self.pos].astype(
                self.data[bp].dtype)

            if self.nChr > 10:  # To avoid crowded labels
                self.ticksLabels = [
                    t if np.mod(int(t), 2)  # Only every two ticks
                    else ''
                    for t in self.data[chrm].unique()
                ]
            else:
                self.ticksLabels = self.data[chrm].unique()  # All the ticks

    def figure(
            self,
            title="Manhattan Plot",
            showgrid=True,
            xlabel=None,
            ylabel='-log10(p)',
            point_size=5,
            showlegend=True,
            col=None,
            suggestiveline_value=-np.log10(1e-8),
            suggestiveline_color='blue',
            suggestiveline_width=1,
            genomewideline_value=-np.log10(5e-8),
            genomewideline_color='red',
            genomewideline_width=1,
            highlight=True,
            highlight_color="red",
    ):
        """

    Keyword arguments:

    - title (string; optional) The title of the graph. (Default:
        "Manhattan Plot")
    - showgrid (bool; optional): Boolean indicating whether gridlines should be
        shown. (Default: True)
    - xlabel (string; optional): Label of the x axis. (Default: None)
    - ylabel: (string; optional): Label of the y axis. (Default:
        "-log10(p)")
    - point_size (number; optional): Size of the points of the Scatter
        plot. (Default: 5)
    - showlegend (bool; optional): Boolean indicating whether legends should be
        shown. (Default: True)
    - col (string; optional): A string representing the color of the
        points of the Scatter plot. Can be in any color format accepted by
        plotly_js graph_objs. (Default: None) Default = None
    - suggestiveline_value (bool/float; optional): A value which must
        be False to deactivate the option, or a numerical value
        corresponding to the p-value at which the line should be drawn.
        The line has no influence on the data points. (Default:
        -np.log10(1e-8))
    - suggestiveline_color (string; optional): Color of the suggestive
      line. (Default: "grey")
    - suggestiveline_width (number): Width of the suggestive
            line. (Default: 2)
    - genomewideline_value (bool/float; optional): A boolean which must be
        False to deactivate the option, or a numerical value corresponding
        to the p-value above which the data points are considered
        significant. (Default: -np.log10(5e-8))
    - genomewideline_color (string; optional): Color of the genome wide
        line. Can be in any color format accepted by plotly_js
        graph_objs. (Default: "red")
    - genomewideline_width (number; optional): Width of the genome wide
      line. (Default: 1)
    - highlight (bool; optional): turning on/off the highlighting of
        data points considered significant. (Default: True)
    - highlight_color (string; optional): Color of the data points
        highlighted because they are significant Can be in any color
        format accepted by plotly_js graph_objs. (Default: "red")

    Returns:
    - A figure formatted for plotly.graph_objs."""

        xmin = min(self.data[self.pos].values)
        xmax = max(self.data[self.pos].values)

        horizontallines = []

        if suggestiveline_value:
            suggestiveline = go.layout.Shape(
                name=SUGGESTIVE_LINE_LABEL,
                type="line",
                fillcolor=suggestiveline_color,
                line=dict(
                    color=suggestiveline_color,
                    width=suggestiveline_width
                ),
                x0=xmin, x1=xmax, xref="x",
                y0=suggestiveline_value, y1=suggestiveline_value, yref="y"
            )
            horizontallines.append(suggestiveline)

        if genomewideline_value:
            genomewideline = go.layout.Shape(
                name=GENOMEWIDE_LINE_LABEL,
                type="line",
                fillcolor=genomewideline_color,
                line=dict(
                    color=genomewideline_color,
                    width=genomewideline_width
                ),
                x0=xmin, x1=xmax, xref="x",
                y0=genomewideline_value, y1=genomewideline_value, yref="y"
            )
            horizontallines.append(genomewideline)

        data_to_plot = []  # To contain the data traces
        tmp = pd.DataFrame()  # Empty DataFrame to contain the highlighted data

        if highlight:
            if not isinstance(highlight, bool):
                if self.snpName not in self.data.columns.values:
                    raise KeyError(
                        "snp argument specified for highlight as %s but "
                        "column not found in the data.frame" % self.snpName
                    )
            else:
                if not genomewideline_value:
                    raise Warning(
                        "The genomewideline_value you entered is not a "
                        "positive value, or False, you cannot set highlight "
                        "to True in that case.")
                tmp = self.data

                # Sort the p-values (or -log10(p-values) above the line
                if genomewideline_value:
                    if self.logp:
                        tmp = tmp.loc[-np.log10(tmp[self.pName])
                                      > genomewideline_value]
                    else:
                        tmp = tmp.loc[tmp[self.pName] > genomewideline_value]

                highlight_hover_text = _get_hover_text(
                    tmp,
                    snpname=self.snpName,
                    genename=self.geneName,
                    annotationname=self.annotationName
                )

                if not tmp.empty:
                    data_to_plot.append(
                        go.Scattergl(
                            x=tmp[self.pos].values,
                            y=-np.log10(tmp[self.pName].values) if self.logp
                            else tmp[self.pName].values,
                            mode="markers",
                            text=highlight_hover_text,
                            marker=dict(
                                color=highlight_color,
                                size=point_size
                            ),
                            name="Point(s) of interest"
                        )
                    )

        # Remove the highlighted data from the DataFrame if not empty
        if tmp.empty:
            data = self.data
        else:
            data = self.data.drop(self.data.index[tmp.index])

        if self.nChr == 1:

            if col is None:
                col = ['black']

            # If single chromosome, ticks and labels automatic.
            layout = go.Layout(
                title=title,
                xaxis={
                    'title': self.xlabel if xlabel is None else xlabel,
                    'showgrid': showgrid,
                    'range': [xmin, xmax],
                },
                yaxis={'title': ylabel},
                hovermode='closest'
            )

            hover_text = _get_hover_text(
                data,
                snpname=self.snpName,
                genename=self.geneName,
                annotationname=self.annotationName
            )

            data_to_plot.append(
                go.Scattergl(
                    x=data[self.pos].values,
                    y=-np.log10(data[self.pName].values) if self.logp
                    else data[self.pName].values,
                    mode="markers",
                    showlegend=showlegend,
                    marker={
                        'color': col[0],
                        'size': point_size,
                        'name': "chr%i" % data[self.chrName].unique()
                    },
                    text=hover_text
                )
            )
        else:
            # if multiple chrms, use the ticks and labels you created above.
            layout = go.Layout(
                title=title,
                xaxis={
                    'title': self.xlabel if xlabel is None else xlabel,
                    'showgrid': showgrid,
                    'range': [xmin, xmax],
                    'tickmode': "array",
                    'tickvals': self.ticks,
                    'ticktext': self.ticksLabels,
                    'ticks': "outside"
                },
                yaxis={'title': ylabel},
                hovermode='closest'
            )

            icol = 0
            if col is None:
                col = [
                    'black' if np.mod(i, 2)
                    else 'grey' for i in range(self.nChr)
                ]

            for i in data[self.index].unique():

                tmp = data[data[self.index] == i]

                chromo = tmp[self.chrName].unique()  # Get chromosome name

                hover_text = _get_hover_text(
                    data,
                    snpname=self.snpName,
                    genename=self.geneName,
                    annotationname=self.annotationName
                )

                data_to_plot.append(
                    go.Scattergl(
                        x=tmp[self.pos].values,
                        y=-np.log10(tmp[self.pName].values) if self.logp
                        else tmp[self.pName].values,
                        mode="markers",
                        showlegend=showlegend,
                        name="Chr%i" % chromo,
                        marker={
                            'color': col[icol],
                            'size': point_size
                        },
                        text=hover_text
                    )
                )

                icol = icol + 1

        layout.shapes = horizontallines

        return go.Figure(data=data_to_plot, layout=layout)
